import CV_Robot.opencv_api as cv_api
from typing import List
import numpy as np
import requests
import base64
import cv2

from CV_Robot import is_Robot, robot_URL, is_Server

net, classes, output_layers = cv_api.load_model()
camera = cv2.VideoCapture()
camera_active = False
is_video = False

class Objects:
    STOP_SIGN = 'STOP_SIGN'
    BIKE = 'BIKE'
    CAR = 'CAR'
    TRAFFIC_LIGHT = "TRAFFIC_LIGHT"
    FIRE_HYDRANT = "FIRE_HYDRANT"
    PERSON = "PERSON"
    TRUCK = "TRUCK"

    map = {"stop sign": STOP_SIGN,
         "bicycle": BIKE,
         "car": CAR,
         "traffic light": TRAFFIC_LIGHT,
         "fire hydrant": FIRE_HYDRANT,
         "person": PERSON,
         "truck": TRUCK}

class VisionObject:
    """
    VisionObject - holds information about a recognized object in an image
    All dimensions are normalized to 100 by 100

    Name: Object - type of object
    Size: float - in square pixels
    BBox: list[float] - x, y, width, height
    X: float - center x
    Y: float - center y
    """
    def __init__(self, box: List[float], conf: float, class_id: int):
        self.Name: Objects = Objects.map[classes[class_id]]
        self.Size: float = round(box[2] * box[3], 3)
        self.BBox = box #x, y, width, height
        self.Conf = round(conf, 3)
        self.X: float = box[0] + box[2] / 2
        self.Y: float = box[1] + box[3] / 2

        self._class_id: int = class_id
        self._class: str = classes[self._class_id]

    def __repr__(self):
        return f"VisionObject<{self.Name}> at ({int(round(self.X))}, {int(round(self.Y))})"

def activate_camera():
    """
    Activates the robot camera
    """
    global camera
    global camera_active
    global is_video
    if not is_Robot:
        camera = cv2.VideoCapture(0, cv2.CAP_DSHOW)
    camera_active = True
    is_video = False

def get_camera_image(skip_frames = 5):
    """
    Retrieves current image from robot camera or video
    Returns none if the video is done or camera failed
    :param skip_frames: If playing a video, skip ahead this many frames at a time
    """
    global camera
    global camera_active
    if not camera_active:
        raise Exception("Camera is not active. Call vision.activate_camera() or vision.load_video()")
    if is_video:
        for i in range(0, skip_frames):
            camera.read()

    if is_Robot:
        #return np.array(json.loads(requests.get('http://' + robot_URL + '/get_camera_array').text)).astype(np.uint8)
        jpg_original = base64.b64decode(requests.get('http://' + robot_URL + '/get_camera_array_fast').text)
        jpg_as_np = np.frombuffer(jpg_original, dtype=np.uint8)
        image_buffer = cv2.imdecode(jpg_as_np, flags=1)
        return image_buffer

    _, img = camera.read()
    return img


def load_image(img_path):
    """
    Loads an image file
    :param img_path: File path to image
    """
    # image loading
    with open(img_path):
        pass #Make sure image exists
    img = cv2.imread(img_path)
    img = cv2.resize(img, None, fx=0.4, fy=0.4)
    return img

def load_video(video_path):
    """
    Loads a video file to emulate camera
    :param video_path: File path to video
    """
    global camera
    global camera_active
    global is_video
    with open(video_path):
        pass #Make sure video exists
    camera = cv2.VideoCapture(video_path)
    camera_active = True
    is_video = True

def get_object_locations(image: cv_api.img_typ, thresh: float = 0.3):
    """
    Returns a list of VisionObjects which can be used to find object location and size
    :param image: Image object (from load_image)
    :param thresh: Threshold to identify object, default is 30% (0.3)
    """
    outputs = cv_api.detect_objects(image, net, output_layers)
    boxes, confs, class_ids = cv_api.get_box_dimensions(outputs, thresh=thresh)
    v_indexes = cv2.dnn.NMSBoxes(boxes, confs, thresh, thresh) #Filter by seperated boxes
    return [VisionObject(*item) for index, item in enumerate(zip(boxes, confs, class_ids))
            if classes[item[2]] in Objects.map.keys() and index in v_indexes]

def show_objects(image: cv_api.img_typ, thresh: float = 0.3, pause: bool = False):
    """
    Displays image with boxes around objects
    :param image: Image object (from load_image)
    :param thresh: Threshold to identify object, default is 30% (0.3)
    :param pause: Set to true to pause after showing image (only applies to local emulation)
    """
    cv_api.draw_labels(get_object_locations(image, thresh=thresh), image, pause=pause)

def find_objects(image: cv_api.img_typ, thresh: float = 0.3):
    """
    Runs machine learning model on image specified, returns list of identified objects
    :param image: Image object (from load_image)
    :param thresh: Threshold to identify object, default is 30% (0.3)
    """
    return list(set(obj.Name for obj in get_object_locations(image, thresh=thresh)))

def show_image(image: cv_api.img_typ, pause: bool = True):
    """
    Displays image
    :param image: Image to display
    :param pause: Set to true to pause after showing image (only applies to local emulation)
    """
    cv_api.show_image(image, pause=pause)

def save_image(image: cv_api.img_typ):
    """
    Saves image to file
    :param image: Image to save
    """
    if is_Robot or is_Server:
        with open('userscripts/s_img.lck', 'w+') as f:
            f.write("lck")
        cv2.imwrite("userscripts/saved_image.png", image)
        with open('userscripts/s_img.lck', 'w+') as f:
            f.write("")
    else:
        cv2.imwrite("saved_image.png", image)


import requests
from time import sleep

try:
    # noinspection PyShadowingBuiltins
    from printlog import printlog as print
except ImportError:
    pass

from CV_Robot import is_Robot, robot_URL

def forward():
    """
    Moves robot forward at speed set by speed command.

    Runs emulated code if GPIO backend not found.
    """
    print("Driving forward...")
    if is_Robot:
        requests.get('http://' + robot_URL + '/forward')

def forward_straight():
    """
    Moves robot forward at speed set by speed command.
    Uses gyro correction to drive straight

    Runs emulated code if GPIO backend not found.
    """
    print("Driving forward (with gyro correction)...")
    if is_Robot:
        requests.get('http://' + robot_URL + '/forward_gyro')


def backward():
    """
    Moves robot backward at speed set by speed command.

    Runs emulated code if GPIO backend not found.
    """
    print("Driving backward...")
    if is_Robot:
        requests.get('http://' + robot_URL + '/backward')

def left():
    """
    Turns robot left at speed set by speed command.

    Runs emulated code if GPIO backend not found.
    """
    print("Turning left...")
    if is_Robot:
        requests.get('http://' + robot_URL + '/left')

def left_angle(x: int):
    """
    Turns robot left to a specified angle at speed set by speed command.
    Runs emulated code if GPIO backend not found.

    :param x: Angle to turn to (0 to 359)
    """
    print(f"Turning left {x} degrees...")
    if is_Robot:
        requests.get('http://' + robot_URL + '/left_angle?angle=' + str(x))

        while requests.get('http://' + robot_URL + '/done').text != "OK":
            sleep(1)


def right():
    """
    Turns robot right at speed set by speed command.

    Runs emulated code if GPIO backend not found.
    """
    print("Turning right...")
    if is_Robot:
        requests.get('http://' + robot_URL + '/right')


def right_angle(x: int):
    """
    Turns robot right to a specified angle at speed set by speed command.
    Runs emulated code if GPIO backend not found.

    :param x: Angle to turn to (0 to 359)
    """
    print(f"Turning right {x} degrees...")
    if is_Robot:
        requests.get('http://' + robot_URL + '/right_angle?angle=' + str(x))

        while requests.get('http://' + robot_URL + '/done').text != "OK":
            sleep(1)

def stop():
    """
    Sets speed of all motors to 0.

    Runs emulated code if GPIO backend not found.
    """
    print("Stopping robot...")
    if is_Robot:
        requests.get('http://' + robot_URL + '/stop')

def speed(val):
    """
    Sets driving speed to value specified
    """
    if is_Robot:
        requests.get('http://' + robot_URL + '/speed?val=' + str(val))


from __future__ import annotations
from typing import TYPE_CHECKING, List
if TYPE_CHECKING:
    from CV_Robot.vision import VisionObject

import matplotlib.pyplot as plt
import numpy as np
import copy
import cv2
import os

from CV_Robot import useLiveImg, is_Server, is_Robot

fig = None
ax = None

img_typ = np.ndarray

def load_model():
    """
    Loads Yolo v3 model (called automtically on import)
    """
    yolov3_path = os.path.dirname(__file__) + '/Models/Yolo v3/'
    try:
        with open(yolov3_path + 'yolov3.weights'):
            pass
    except FileNotFoundError:
        print("Downloading model...")
        import wget
        import sys
        def bar_custom(current, total, _): #_ = width
            sys.stdout.write("\r")
            sys.stdout.write("Downloading: %d%% [%d / %d] bytes" % (current / total * 100, current, total))

        wget.download('https://pjreddie.com/media/files/yolov3.weights', yolov3_path + 'yolov3.weights', bar=bar_custom)
        print()

    net: cv2.dnn_Net = cv2.dnn.readNet(yolov3_path + "yolov3.weights", yolov3_path + "yolov3.cfg")
    with open(yolov3_path + "yolov3.names", "r") as f:
        classes = [line.strip() for line in f.readlines()]

    output_layers = [layer_name for layer_name in net.getUnconnectedOutLayersNames()]
    return net, classes, output_layers

cache = []
def detect_objects(img: img_typ, net: cv2.dnn_Net, outputLayers: list):
    """
    Returns DNN outputs for finding objects in image, uses cache if possible
    """
    retval = None
    for c_img, c_outputs in cache:
        if np.array_equal(img, c_img):
            retval = c_outputs

    if retval is None:
        blob = cv2.dnn.blobFromImage(img, scalefactor=0.00392, size=(320, 320), mean=(0, 0, 0), swapRB=True, crop=False)
        net.setInput(blob)
        retval = net.forward(outputLayers)

    if len(cache) >= 5:
        cache.pop(0)
    cache.append((img, retval))

    return retval

def get_box_dimensions(outputs: list, thresh: float = 0.3):
    """
    Returns X, Y, width, height of objects
    """
    boxes = []
    confs = []
    class_ids = []
    for output in outputs:
        for detect in output:
            scores = detect[5:]
            class_id = np.argmax(scores)
            conf = scores[class_id]
            if conf > thresh:
                center_x = detect[0] * 100
                center_y = detect[1] * 100
                w = round(detect[2] * 100, 3)
                h = round(detect[3] * 100, 3)
                x = round(center_x - w / 2, 3)
                y = round(center_y - h / 2, 3)
                boxes.append([x, y, w, h])
                confs.append(float(conf))
                class_ids.append(class_id)
    return boxes, confs, class_ids

def show_image(img: img_typ, pause: bool =False):
    """
    Displays Matplotlib render of image
    If on Colab - display in new figure
    If on Server - display in exisiting figure with optional pause
    If on Robot - save as "cv_robot_img.py"
    """
    global fig
    global ax

    if is_Robot or (is_Server and not useLiveImg):
        with open('userscripts/img.lck', 'w+') as f:
            f.write("lck")
        cv2.imwrite("userscripts/cv_robot_img.png", img)
        with open('userscripts/img.lck', 'w+') as f:
            f.write("")
        return

    t_img = img[:, :, ::-1]  # convert BGR (for opencv) to RGB (for matplotlib)
    if fig is None or not plt.fignum_exists(fig.number):
        fig = plt.figure(figsize=(10, 10))
        ax = fig.add_subplot()
    ax.clear()
    ax.imshow(t_img)

    x_space = img.shape[1] * 0.25
    y_space = img.shape[0] * 0.25

    ax.set_xticks([i * x_space for i in range(0, 5)], minor=False)
    ax.set_yticks([i * y_space for i in range(0, 5)], minor=False)
    ax.set_xticklabels(list(range(0, 101, 25)), fontdict=None, minor=False)
    ax.set_yticklabels(list(range(0, 101, 25)), fontdict=None, minor=False)

    if is_Server and not pause:
        plt.draw()
        plt.pause(0.1)
    else:
        plt.show()



def draw_labels(objects: List[VisionObject], img: img_typ, pause: bool = False):
    """
    Draw labeled boxes on image
    """
    global fig
    global ax

    img = copy.deepcopy(img)
    font = cv2.FONT_HERSHEY_PLAIN
    for index, obj in enumerate(objects):
        x, y, w, h = obj.BBox
        x = int(x * img.shape[1] / 100)
        w = int(w * img.shape[1] / 100)
        y = int(y * img.shape[0] / 100)
        h = int(h * img.shape[0] / 100)

        label = str(obj.Name)
        color = (0,0,255) #red
        cv2.rectangle(img, (x, y), (x + w, y + h), color, 2)
        cv2.putText(img, label, (x, y - 5), font, 1, color, 1)

    show_image(img, pause=pause)





try:
    # noinspection PyShadowingBuiltins
    from printlog import printlog as print
except ImportError:
    pass

try:
    import CONFIG_CV_SERVER
    is_Server = True
except ImportError:
    is_Server = False

try:
    import CONFIG_CV_LIVE_IMG
    useLiveImg = True
except ImportError:
    useLiveImg = False

try:
    from CONFIG_CV_ROBOT import robot_URL
    is_Robot = True
except ImportError:
    is_Robot = False
    robot_URL = "0.0.0.0"

print("CV Robot version 0.4.3")
if is_Robot:
    print("Running on robot...")
elif is_Server:
    print("Running on server...")
else:
    print("Running in emulation mode...")